const express = require('express');

const libs = `${process.cwd()}/libs`;

const squel = require('squel');

const log = require(`${libs}/log`)(module);

const query = require(`${libs}/middlewares/query`);

const sqlQuery = require(`${libs}/db/query_exec`);

const response = require(`${libs}/middlewares/response`);

const locationApp = express();

function locationIdToStr(locationId) {
    let locationStr = 'Total';
    switch(locationId) {
        case 1:
            locationStr = 'Urbana';                         break;
        case 2:
            locationStr = 'Rural';                          break;
        case 3:
            locationStr = 'Área de assentamento';           break;
        case 4:
            locationStr = 'Terra indígena';                 break;
        case 5:
            locationStr = 'Área remanescente de quilombos'; break;
        case 6:
            locationStr = 'Unidade de uso sustentável';     break;

    }
    return locationStr;
}

function schoolYearIdToStr(schoolYearId)
{
    let schoolYearStr;
    switch(schoolYearId) {
        case 11:
            schoolYearStr = 'Creche';
            break;
        case 21:
            schoolYearStr = 'Pré-escola';
            break;
        case 31:
            schoolYearStr = '1 Ano';
            break;
        case 32:
            schoolYearStr = '2 Ano - 1 Serie';
            break;
        case 33:
            schoolYearStr = '3 Ano - 2 Serie';
            break;
        case 34:
            schoolYearStr = '4 Ano - 3 Serie';
            break;
        case 35:
            schoolYearStr = '5 Ano - 4 Serie';
            break;
        case 41:
            schoolYearStr = '6 Ano - 5 Serie';
            break;
        case 42:
            schoolYearStr = '7 Ano - 6 Serie';
            break;
        case 43:
            schoolYearStr = '8 Ano - 7 Serie';
            break;
        case 44:
            schoolYearStr = '9 Ano - 8 Serie';
            break;
        case 51:
            schoolYearStr = '1 Ano'; // equivalent to 'EM 1 Série'
            break;
        case 52:
            schoolYearStr = '2 Ano'; // equivalent to 'EM 2 Série'
            break;
        case 53:
            schoolYearStr = '3 Ano'; // equivalent to 'EM 3 Série'
            break;
        case 54:
            schoolYearStr = '4 Ano'; // equivalent to 'EM 4 Série'
            break;
        case 61:
            schoolYearStr = 'EJA AI';
            break;
        case 62:
            schoolYearStr = 'EJA AF';
            break;
        case 63:
            schoolYearStr = 'EJA EM';
            break;
        case 64:
            schoolYearStr = 'EJA semi-presencial';
            break;
        case 71:
            schoolYearStr = 'EP';
            break;
        case 81:
            schoolYearStr = 'Atividades complementares e AEE';
            break;
        default:
            schoolYearStr = 'Não classificado';
    }
    return schoolYearStr;
}

function processResultSet(querySet, querySetLabels = ["result"], singleResult = false) {
    const resultMap = new Map();
    let resultIdx = 0;
    // loop relies on the fact that Promise.all maintains the order of the original iterable
    for(let result of querySet) {
        const resultLbl = querySetLabels[resultIdx];
        resultMap[resultLbl] = [];
        if (singleResult) {
            resultMap[resultLbl] = result[0];
        } else {
            for(let row of result) {
                log.debug(row);
                resultMap[resultLbl].push(row);
            }
        }
        resultIdx++;
    }
    log.debug(resultMap);
    return resultMap;
}

function dbExecAll(querySet = []) {
    // Issue all queries concurrently to the database, for every query object in the iterable
    // NOTE: Array.map() returns a copy of the original array with each object 'mapped'.
    return querySet.map((qry) => { return sqlQuery(qry.toString()); });
}

locationApp.get('/sociodemographic', (req, res, next) => {
    const populationYearQry = squel.select()
        .field('MAX(ibge_populacao.ano_censo)')
        .from('ibge_populacao');

    const populationQry = squel.select()
        .field('\'Brasil\'', 'name')
        .field('SUM(populacao)', 'population')
        .field('ibge_populacao.ano_censo', 'census_year')
        .from('ibge_populacao')
        .where(`ibge_populacao.ano_censo IN (${populationYearQry.toString()})`)
        .group('ibge_populacao.ano_censo');

    const pibYearQry = squel.select()
        .field('MAX(ibge_pib.ano_censo)')
        .from('ibge_pib');

    const pibQry = squel.select()
        .field('\'Brasil\'', 'name')
        .field('AVG(ibge_pib.pib_per_capita)', 'gdp_per_capita')
        .field('ibge_pib.ano_censo', 'census_year')
        .from('ibge_pib')
        .where(`ibge_pib.ano_censo IN (${pibYearQry.toString()})`)
        .group('ibge_pib.ano_censo');

    const idhYearQry = squel.select()
        .field('MAX(adh_idh.ano_censo)')
        .from('adh_idh');

    const idhQry = squel.select()
        .field('\'Brasil\'', 'name')
        .field('AVG(idhm)', 'idhm')
        .field('adh_idh.ano_censo', 'census_year')
        .from('adh_idh')
        .where(`adh_idh.ano_censo IN (${idhYearQry.toString()})`)
        .group('adh_idh.ano_censo');

    const analfabYearQry = squel.select()
        .field('MAX(adh_analfabetismo.ano_censo)')
        .from('adh_analfabetismo');

    const analfabQry = squel.select()
        .field('\'Brasil\'', 'name')
        .field('AVG(t_analf15m)', 'analfabetism')
        .field('adh_analfabetismo.ano_censo', 'census_year')
        .from('adh_analfabetismo')
        .where(`adh_analfabetismo.ano_censo IN (${analfabYearQry.toString()})`)
        .group('adh_analfabetismo.ano_censo');

    const giniYearQry = squel.select()
        .field('MAX(adh_gini.ano_censo)')
        .from('adh_gini');

    const giniQry = squel.select()
        .field('\'Brasil\'', 'name')
        .field('AVG(gini)', 'gini')
        .field('adh_gini.ano_censo', 'census_year')
        .from('adh_gini')
        .where(`adh_gini.ano_censo IN (${giniYearQry.toString()})`)
        .group('adh_gini.ano_censo');

    // map query objects to their respective response labels
    const queryLabels = [ "population", "gdp", "idh", "analfab", "gini" ];
    const querySet = [ populationQry, pibQry, idhQry, analfabQry, giniQry ];
    // wait until all queries finish or one of them fail
    Promise.all(dbExecAll(querySet)).then((queryResults) => {
        req.result = processResultSet(queryResults, queryLabels, true);
        next();
    }).catch((error) => {
        log.error(`[SQL query error] ${error}`);
        next(error);
    });
}, response('location'));

locationApp.get('/sociodemographic/region/:id', (req, res, next) => {
    const regionId = parseInt(req.params.id, 10);

    const populationYearQry = squel.select()
        .field('MAX(ibge_populacao.ano_censo)')
        .from('ibge_populacao');

    const populationQry = squel.select()
        .field('regiao.nome', 'name')
        .field('SUM(populacao)', 'population')
        .field('ano_censo', 'census_year')
        .from('ibge_populacao')
        .from('municipio')
        .from('estado')
        .from('regiao')
        .where(`regiao.pk_regiao_id = ${regionId}`)
        .where(`ibge_populacao.ano_censo IN (${populationYearQry.toString()})`)
        .where('ibge_populacao.fk_municipio_id = municipio.pk_cod_ibge')
        .where('municipio.fk_estado_id = estado.pk_estado_id')
        .where('estado.fk_regiao_id = regiao.pk_regiao_id')
        .group('regiao.nome')
        .group('ibge_populacao.ano_censo')
        .order('regiao.nome');

    const pibYearQry = squel.select()
        .field('MAX(ibge_pib.ano_censo)')
        .from('ibge_pib');

    const pibQry = squel.select()
        .field('regiao.nome', 'name')
        .field('AVG(ibge_pib.pib_per_capita)', 'gdp_per_capita')
        .field('ano_censo', 'census_year')
        .from('ibge_pib')
        .from('municipio')
        .from('estado')
        .from('regiao')
        .where(`regiao.pk_regiao_id = ${regionId}`)
        .where(`ibge_pib.ano_censo IN (${pibYearQry.toString()})`)
        .where('ibge_pib.fk_municipio_id = municipio.pk_cod_ibge')
        .where('municipio.fk_estado_id = estado.pk_estado_id')
        .where('estado.fk_regiao_id = regiao.pk_regiao_id')
        .group('regiao.nome')
        .group('ibge_pib.ano_censo')
        .order('regiao.nome');

    const idhYearQry = squel.select()
        .field('MAX(adh_idh.ano_censo)')
        .from('adh_idh');

    const idhQry = squel.select()
        .field('regiao.nome', 'name')
        .field('AVG(idhm)', 'idhm')
        .field('ano_censo', 'census_year')
        .from('adh_idh')
        .from('municipio')
        .from('estado')
        .from('regiao')
        .where(`regiao.pk_regiao_id = ${regionId}`)
        .where(`adh_idh.ano_censo IN (${idhYearQry.toString()})`)
        .where('adh_idh.fk_municipio_id = municipio.pk_cod_ibge')
        .where('municipio.fk_estado_id = estado.pk_estado_id')
        .where('estado.fk_regiao_id = regiao.pk_regiao_id')
        .group('regiao.nome')
        .group('adh_idh.ano_censo')
        .order('regiao.nome');

    const analfabYearQry = squel.select()
        .field('MAX(adh_analfabetismo.ano_censo)')
        .from('adh_analfabetismo');

    const analfabQry = squel.select()
        .field('regiao.nome', 'name')
        .field('AVG(t_analf15m)', 'analfabetism')
        .field('ano_censo', 'census_year')
        .from('adh_analfabetismo')
        .from('municipio')
        .from('estado')
        .from('regiao')
        .where(`regiao.pk_regiao_id = ${regionId}`)
        .where(`adh_analfabetismo.ano_censo IN (${analfabYearQry.toString()})`)
        .where('adh_analfabetismo.fk_municipio_id = municipio.pk_cod_ibge')
        .where('municipio.fk_estado_id = estado.pk_estado_id')
        .where('estado.fk_regiao_id = regiao.pk_regiao_id')
        .group('regiao.nome')
        .group('adh_analfabetismo.ano_censo')
        .order('regiao.nome');

    const giniYearQry = squel.select()
        .field('MAX(adh_gini.ano_censo)')
        .from('adh_gini');

    const giniQry = squel.select()
        .field('regiao.nome', 'name')
        .field('AVG(gini)', 'gini')
        .field('ano_censo', 'census_year')
        .from('adh_gini')
        .from('municipio')
        .from('estado')
        .from('regiao')
        .where(`regiao.pk_regiao_id = ${regionId}`)
        .where(`adh_gini.ano_censo IN (${giniYearQry.toString()})`)
        .where('adh_gini.fk_municipio_id = municipio.pk_cod_ibge')
        .where('municipio.fk_estado_id = estado.pk_estado_id')
        .where('estado.fk_regiao_id = regiao.pk_regiao_id')
        .group('regiao.nome')
        .group('adh_gini.ano_censo')
        .order('regiao.nome');
    // map query objects to their respective response labels
    const queryLabels = [ "population", "gdp", "idh", "analfab", "gini" ];
    const querySet = [ populationQry, pibQry, idhQry, analfabQry, giniQry ];
    // wait until all queries finish or one of them fail
    Promise.all(dbExecAll(querySet)).then((queryResults) => {
        req.result = processResultSet(queryResults, queryLabels, true);
        next();
    }).catch((error) => {
        log.error(`[SQL query error] ${error}`);
        next(error);
    });
}, response('location'));

locationApp.get('/sociodemographic/state/:id', (req, res, next) => {
    const stateId = parseInt(req.params.id, 10);

    const populationYearQry = squel.select()
        .field('MAX(ibge_populacao.ano_censo)')
        .from('ibge_populacao');
    // load all cities by state and compute the sociodemographic and educational information
    const populationQry = squel.select()
        .field('estado.nome', 'name')
        .field('SUM(populacao)', 'population')
        .field('ibge_populacao.ano_censo', 'census_year')
        .from('ibge_populacao')
        .from('municipio')
        .from('estado')
        .where(`estado.pk_estado_id = ${stateId}`)
        .where(`ibge_populacao.ano_censo IN (${populationYearQry.toString()})`)
        .where('ibge_populacao.fk_municipio_id = municipio.pk_cod_ibge')
        .where('municipio.fk_estado_id = estado.pk_estado_id')
        .group('estado.nome')
        .group('ibge_populacao.ano_censo')
        .order('estado.nome');

    const pibYearQry = squel.select()
        .field('MAX(ibge_pib.ano_censo)')
        .from('ibge_pib');

    const pibQry = squel.select()
        .field('estado.nome', 'name')
        .field('AVG(ibge_pib.pib_per_capita)', 'gdp_per_capita')
        .field('ibge_pib.ano_censo', 'census_year')
        .from('ibge_pib')
        .from('municipio')
        .from('estado')
        .where(`estado.pk_estado_id = ${stateId}`)
        .where(`ibge_pib.ano_censo IN (${pibYearQry.toString()})`)
        .where('ibge_pib.fk_municipio_id = municipio.pk_cod_ibge')
        .where('municipio.fk_estado_id = estado.pk_estado_id')
        .group('estado.nome')
        .group('ibge_pib.ano_censo')
        .order('estado.nome');

    const idhYearQry = squel.select()
        .field('MAX(adh_idh.ano_censo)')
        .from('adh_idh');

    const idhQry = squel.select()
        .field('estado.nome', 'name')
        .field('AVG(idhm)', 'idhm')
        .field('adh_idh.ano_censo', 'census_year')
        .from('adh_idh')
        .from('municipio')
        .from('estado')
        .where(`estado.pk_estado_id = ${stateId}`)
        .where(`adh_idh.ano_censo IN (${idhYearQry.toString()})`)
        .where('adh_idh.fk_municipio_id = municipio.pk_cod_ibge')
        .where('municipio.fk_estado_id = estado.pk_estado_id')
        .group('estado.nome')
        .group('adh_idh.ano_censo')
        .order('estado.nome');

    const analfabYearQry = squel.select()
        .field('MAX(adh_analfabetismo.ano_censo)')
        .from('adh_analfabetismo');

    const analfabQry = squel.select()
        .field('estado.nome', 'name')
        .field('AVG(t_analf15m)', 'analfabetism')
        .field('adh_analfabetismo.ano_censo', 'census_year')
        .from('adh_analfabetismo')
        .from('municipio')
        .from('estado')
        .where(`estado.pk_estado_id = ${stateId}`)
        .where(`adh_analfabetismo.ano_censo IN (${analfabYearQry.toString()})`)
        .where('adh_analfabetismo.fk_municipio_id = municipio.pk_cod_ibge')
        .where('municipio.fk_estado_id = estado.pk_estado_id')
        .group('estado.nome')
        .group('adh_analfabetismo.ano_censo')
        .order('estado.nome');

    const giniYearQry = squel.select()
        .field('MAX(adh_gini.ano_censo)')
        .from('adh_gini');

    const giniQry = squel.select()
        .field('estado.nome', 'name')
        .field('AVG(gini)', 'gini')
        .field('adh_gini.ano_censo', 'census_year')
        .from('adh_gini')
        .from('municipio')
        .from('estado')
        .where(`estado.pk_estado_id = ${stateId}`)
        .where(`adh_gini.ano_censo IN (${giniYearQry.toString()})`)
        .where('adh_gini.fk_municipio_id = municipio.pk_cod_ibge')
        .where('municipio.fk_estado_id = estado.pk_estado_id')
        .group('estado.nome')
        .group('adh_gini.ano_censo')
        .order('estado.nome');
    // map query objects to their respective response labels
    const queryLabels = [ "population", "gdp", "idh", "analfab", "gini" ];
    const querySet = [ populationQry, pibQry, idhQry, analfabQry, giniQry ];
    // wait until all queries finish or one of them fail
    Promise.all(dbExecAll(querySet)).then((queryResults) => {
        req.result = processResultSet(queryResults, queryLabels, true);
        next();
    }).catch((error) => {
        log.error(`[SQL query error] ${error}`);
        next(error);
    });
}, response('location'));

locationApp.get('/sociodemographic/city/:id', (req, res, next) => {
    const cityId = parseInt(req.params.id, 10);

    const populationYearQry = squel.select()
        .field('MAX(ibge_populacao.ano_censo)')
        .from('ibge_populacao');
    // load all cities by state and compute the sociodemographic and educational information
    const populationQry = squel.select()
        .field('municipio.nome', 'name')
        .field('SUM(populacao)', 'population')
        .field('ibge_populacao.ano_censo', 'census_year')
        .from('ibge_populacao')
        .from('municipio')
        .where(`municipio.pk_cod_ibge = ${cityId}`)
        .where(`ibge_populacao.ano_censo IN (${populationYearQry.toString()})`)
        .where('ibge_populacao.fk_municipio_id = municipio.pk_cod_ibge')
        .group('municipio.nome')
        .group('ibge_populacao.ano_censo')
        .order('municipio.nome');

    const pibYearQry = squel.select()
        .field('MAX(ibge_pib.ano_censo)')
        .from('ibge_pib');

    const pibQry = squel.select()
        .field('municipio.nome', 'name')
        .field('AVG(ibge_pib.pib_per_capita)', 'gdp_per_capita')
        .field('ibge_pib.ano_censo', 'census_year')
        .from('ibge_pib')
        .from('municipio')
        .where(`municipio.pk_cod_ibge = ${cityId}`)
        .where(`ibge_pib.ano_censo IN (${pibYearQry.toString()})`)
        .where('ibge_pib.fk_municipio_id = municipio.pk_cod_ibge')
        .group('municipio.nome')
        .group('ibge_pib.ano_censo')
        .order('municipio.nome');

    const idhYearQry = squel.select()
        .field('MAX(adh_idh.ano_censo)')
        .from('adh_idh');

    const idhQry = squel.select()
        .field('municipio.nome', 'name')
        .field('AVG(idhm)', 'idhm')
        .field('adh_idh.ano_censo', 'census_year')
        .from('adh_idh')
        .from('municipio')
        .where(`municipio.pk_cod_ibge = ${cityId}`)
        .where(`adh_idh.ano_censo IN (${idhYearQry.toString()})`)
        .where('adh_idh.fk_municipio_id = municipio.pk_cod_ibge')
        .group('municipio.nome')
        .group('adh_idh.ano_censo')
        .order('municipio.nome');

    const analfabYearQry = squel.select()
        .field('MAX(adh_analfabetismo.ano_censo)')
        .from('adh_analfabetismo');

    const analfabQry = squel.select()
        .field('municipio.nome', 'name')
        .field('AVG(t_analf15m)', 'analfabetism')
        .field('adh_analfabetismo.ano_censo', 'census_year')
        .from('adh_analfabetismo')
        .from('municipio')
        .where(`municipio.pk_cod_ibge = ${cityId}`)
        .where(`adh_analfabetismo.ano_censo IN (${analfabYearQry.toString()})`)
        .where('adh_analfabetismo.fk_municipio_id = municipio.pk_cod_ibge')
        .group('municipio.nome')
        .group('adh_analfabetismo.ano_censo')
        .order('municipio.nome');

    const giniYearQry = squel.select()
        .field('MAX(adh_gini.ano_censo)')
        .from('adh_gini');

    const giniQry = squel.select()
        .field('municipio.nome', 'name')
        .field('AVG(gini)', 'gini')
        .field('adh_gini.ano_censo', 'census_year')
        .from('adh_gini')
        .from('municipio')
        .where(`municipio.pk_cod_ibge = ${cityId}`)
        .where(`adh_gini.ano_censo IN (${giniYearQry.toString()})`)
        .where('adh_gini.fk_municipio_id = municipio.pk_cod_ibge')
        .group('municipio.nome')
        .group('adh_gini.ano_censo')
        .order('municipio.nome');
    // map query objects to their respective response labels
    const queryLabels = [ "population", "gdp", "idh", "analfab", "gini" ];
    const querySet = [ populationQry, pibQry, idhQry, analfabQry, giniQry ];
    // wait until all queries finish or one of them fail
    Promise.all(dbExecAll(querySet)).then((queryResults) => {
        req.result = processResultSet(queryResults, queryLabels, true);
        next();
    }).catch((error) => {
        log.error(`[SQL query error] ${error}`);
        next(error);
    });
}, response('location'));

locationApp.get('/educational', (req, res, next) => {
    const censusYearQry = squel.select()
        .field('MAX(escola.ano_censo)', 'ano_censo')
        .from('escola')
        .toString();

    const totalSchoolsQry = squel.select()
        .field('\'Brasil\'', 'name')
        .field('0', 'location')
        .field('COUNT(DISTINCT(escola.cod_entidade))', 'total')
        .field('escola.ano_censo', 'census_year')
        .from('escola')
        .where(`escola.ano_censo IN (${censusYearQry})`)
        .where('escola.id_tipo_turma = 0')
        .group('escola.ano_censo');

    const schoolsPerLocationQry = squel.select()
        .field('\'Brasil\'', 'name')
        .field('escola.tipo_localizacao', 'location')
        .field('COUNT(DISTINCT(escola.cod_entidade))', 'total')
        .field('escola.ano_censo', 'census_year')
        .from('escola')
        .where(`escola.ano_censo IN (${censusYearQry})`)
        .where('escola.id_tipo_turma = 0')
        .group('escola.tipo_localizacao')
        .group('escola.ano_censo')
        .order('escola.tipo_localizacao');

    const schoolClassYearQry = squel.select()
        .field('MAX(turma.ano_censo)')
        .from('turma')
        .toString();

    const enrollmentsQry = squel.select()
        .field('\'Brasil\'', 'name')
        .field('COALESCE(SUM(turma.num_matriculas), 0)', 'total')
        .field('turma.ano_censo', 'census_year')
        .from('turma')
        .where(`turma.ano_censo IN (${schoolClassYearQry})`)
        .where('turma.fk_tipo_turma_id <= 3')
        .group('turma.ano_censo');

    const enrollmentsPerAdmDepQry = squel.select()
        .field('\'Brasil\'', 'name')
        .field('COALESCE(SUM(turma.num_matriculas), 0)', 'total')
        .field('turma.ano_censo', 'census_year')
        .field('dependencia_adm.nome', 'adm_dependency')
        .from('turma')
        .from('dependencia_adm')
        .where('turma.fk_dependencia_adm_id = dependencia_adm.pk_dependencia_adm_id')
        .where('turma.fk_tipo_turma_id <= 3')
        .where(`turma.ano_censo IN (${schoolClassYearQry})`)
        .group('turma.ano_censo')
        .group('dependencia_adm.nome')
        .order('dependencia_adm.nome');

    const enrollmentsPerSchoolLevelQry = squel.select()
        .field('\'Brasil\'', 'name')
        .field('COALESCE(SUM(turma.num_matriculas), 0)', 'total')
        .field('turma.ano_censo', 'census_year')
        .field('etapa_ensino.desc_etapa', 'school_level')
        .from('turma')
        .from('etapa_ensino')
        .where('turma.fk_etapa_ensino_id = etapa_ensino.pk_etapa_ensino_id')
        .where('turma.fk_tipo_turma_id <= 3')
        .where(`turma.ano_censo IN (${schoolClassYearQry})`)
        .group('turma.ano_censo')
        .group('etapa_ensino.desc_etapa')
        .group('etapa_ensino.pk_etapa_ensino_id')
        .order('etapa_ensino.pk_etapa_ensino_id');

    const queryLabels = [ "school", "school_per_location", "enrollment", "enrollment_per_adm_dep",
        "enrollment_per_school_level" ];
    const querySet = [ totalSchoolsQry, schoolsPerLocationQry, enrollmentsQry,
        enrollmentsPerAdmDepQry, enrollmentsPerSchoolLevelQry];
    // wait until all queries finish or one of them fail
    Promise.all(dbExecAll(querySet)).then((queryResults) => {
        req.result = processResultSet(queryResults, queryLabels);
        for(let label in req.result) {
            for(let row of req.result[label]) {
                if (row.hasOwnProperty('location')) {
                    row.location = locationIdToStr(row.location);
                }
            }
        }
        next();
    }).catch((error) => {
        log.error(`[SQL query error] ${error}`);
        next(error);
    });
}, response('location'));

locationApp.get('/educational/region/:id', (req, res, next) => {
    const regionId = parseInt(req.params.id, 10);

    const censusYearQry = squel.select()
        .field('MAX(escola.ano_censo)', 'ano_censo')
        .from('escola')
        .toString();

    const totalSchoolsQry = squel.select()
        .field('regiao.nome', 'name')
        .field('0', 'location')
        .field('COUNT(DISTINCT(escola.cod_entidade))', 'total')
        .field('escola.ano_censo', 'census_year')
        .from('escola')
        .from('estado')
        .from('regiao')
        .where('escola.fk_estado_id = estado.pk_estado_id')
        .where('estado.fk_regiao_id = regiao.pk_regiao_id')
        .where(`regiao.pk_regiao_id = ${regionId}`)
        .where(`escola.ano_censo IN (${censusYearQry})`)
        .where('escola.id_tipo_turma = 0')
        .group('regiao.nome')
        .group('escola.ano_censo')
        .order('regiao.nome');

    const schoolsPerLocationQry = squel.select()
        .field('regiao.nome', 'name')
        .field('escola.tipo_localizacao', 'location')
        .field('COUNT(DISTINCT(escola.cod_entidade))', 'total')
        .field('escola.ano_censo', 'census_year')
        .from('escola')
        .from('estado')
        .from('regiao')
        .where('escola.fk_estado_id = estado.pk_estado_id')
        .where('estado.fk_regiao_id = regiao.pk_regiao_id')
        .where(`regiao.pk_regiao_id = ${regionId}`)
        .where(`escola.ano_censo IN (${censusYearQry})`)
        .where('escola.id_tipo_turma = 0')
        .group('regiao.nome')
        .group('escola.tipo_localizacao')
        .group('escola.ano_censo')
        .order('regiao.nome')
        .order('escola.tipo_localizacao');

    const schoolClassYearQry = squel.select()
        .field('MAX(turma.ano_censo)')
        .from('turma')
        .toString();

    const enrollmentsQry = squel.select()
        .field('regiao.nome', 'name')
        .field('COALESCE(SUM(turma.num_matriculas), 0)', 'total')
        .field('turma.ano_censo', 'census_year')
        .from('turma')
        .from('regiao')
        .where('turma.fk_regiao_id = regiao.pk_regiao_id')
        .where(`turma.fk_regiao_id = ${regionId}`)
        .where(`turma.ano_censo IN (${schoolClassYearQry})`)
        .where('turma.fk_tipo_turma_id <= 3')
        .group('turma.ano_censo')
        .group('regiao.nome')
        .order('regiao.nome');

    const enrollmentsPerAdmDepQry = squel.select()
        .field('regiao.nome', 'name')
        .field('COALESCE(SUM(turma.num_matriculas), 0)', 'total')
        .field('turma.ano_censo', 'census_year')
        .field('dependencia_adm.nome', 'adm_dependency')
        .from('turma')
        .from('dependencia_adm')
        .from('regiao')
        .where('turma.fk_regiao_id = regiao.pk_regiao_id')
        .where(`turma.fk_regiao_id = ${regionId}`)
        .where('turma.fk_dependencia_adm_id = dependencia_adm.pk_dependencia_adm_id')
        .where('turma.fk_tipo_turma_id <= 3')
        .where(`turma.ano_censo IN (${schoolClassYearQry})`)
        .group('turma.ano_censo')
        .group('dependencia_adm.nome')
        .group('regiao.nome')
        .order('regiao.nome')
        .order('dependencia_adm.nome');

    const enrollmentsPerSchoolLevelQry = squel.select()
        .field('regiao.nome', 'name')
        .field('COALESCE(SUM(turma.num_matriculas), 0)', 'total')
        .field('turma.ano_censo', 'census_year')
        .field('etapa_ensino.desc_etapa', 'school_level')
        .from('turma')
        .from('etapa_ensino')
        .from('regiao')
        .where('turma.fk_regiao_id = regiao.pk_regiao_id')
        .where(`turma.fk_regiao_id = ${regionId}`)
        .where('turma.fk_etapa_ensino_id = etapa_ensino.pk_etapa_ensino_id')
        .where('turma.fk_tipo_turma_id <= 3')
        .where(`turma.ano_censo IN (${schoolClassYearQry})`)
        .group('turma.ano_censo')
        .group('etapa_ensino.desc_etapa')
        .group('regiao.nome')
        .group('etapa_ensino.pk_etapa_ensino_id')
        .order('regiao.nome')
        .order('etapa_ensino.pk_etapa_ensino_id');

    const queryLabels = [ "school", "school_per_location", "enrollment", "enrollment_per_adm_dep",
        "enrollment_per_school_level" ];
    const querySet = [ totalSchoolsQry, schoolsPerLocationQry, enrollmentsQry,
        enrollmentsPerAdmDepQry, enrollmentsPerSchoolLevelQry];
    // wait until all queries finish or one of them fail
    Promise.all(dbExecAll(querySet)).then((queryResults) => {
        req.result = processResultSet(queryResults, queryLabels);
        for(let label in req.result) {
            for(let row of req.result[label]) {
                if (row.hasOwnProperty('location')) {
                    row.location = locationIdToStr(row.location);
                }
            }
        }
        next();
    }).catch((error) => {
        log.error(`[SQL query error] ${error}`);
        next(error);
    });
}, response('location'));

locationApp.get('/educational/state/:id', (req, res, next) => {
    const stateId = parseInt(req.params.id, 10);

    const censusYearQry = squel.select()
        .field('MAX(escola.ano_censo)', 'ano_censo')
        .from('escola')
        .toString();

    const totalSchoolsQry = squel.select()
        .field('estado.nome', 'name')
        .field('0', 'location')
        .field('COUNT(DISTINCT(escola.cod_entidade))', 'total')
        .field('escola.ano_censo', 'census_year')
        .from('escola')
        .from('estado')
        .where('escola.fk_estado_id = estado.pk_estado_id')
        .where(`escola.fk_estado_id = ${stateId}`)
        .where(`escola.ano_censo IN (${censusYearQry})`)
        .where('escola.id_tipo_turma = 0')
        .group('estado.nome')
        .group('escola.ano_censo')
        .order('estado.nome');

    const schoolsPerLocationQry = squel.select()
        .field('estado.nome', 'name')
        .field('escola.tipo_localizacao', 'location')
        .field('COUNT(DISTINCT(escola.cod_entidade))', 'total')
        .field('escola.ano_censo', 'census_year')
        .from('escola')
        .from('estado')
        .where('escola.fk_estado_id = estado.pk_estado_id')
        .where(`escola.fk_estado_id = ${stateId}`)
        .where(`escola.ano_censo IN (${censusYearQry})`)
        .where('escola.id_tipo_turma = 0')
        .group('estado.nome')
        .group('escola.tipo_localizacao')
        .group('escola.ano_censo')
        .order('estado.nome')
        .order('escola.tipo_localizacao');

    const schoolClassYearQry = squel.select()
        .field('MAX(turma.ano_censo)')
        .from('turma')
        .toString();

    const enrollmentsQry = squel.select()
        .field('estado.nome', 'name')
        .field('COALESCE(SUM(turma.num_matriculas), 0)', 'total')
        .field('turma.ano_censo', 'census_year')
        .from('turma')
        .from('estado')
        .where('turma.fk_estado_id = estado.pk_estado_id')
        .where(`turma.fk_estado_id = ${stateId}`)
        .where(`turma.ano_censo IN (${schoolClassYearQry})`)
        .where('turma.fk_tipo_turma_id <= 3')
        .group('turma.ano_censo')
        .group('estado.nome')
        .order('estado.nome');

    const enrollmentsPerAdmDepQry = squel.select()
        .field('estado.nome', 'name')
        .field('COALESCE(SUM(turma.num_matriculas), 0)', 'total')
        .field('turma.ano_censo', 'census_year')
        .field('dependencia_adm.nome', 'adm_dependency')
        .from('turma')
        .from('dependencia_adm')
        .from('estado')
        .where('turma.fk_estado_id = estado.pk_estado_id')
        .where(`turma.fk_estado_id = ${stateId}`)
        .where('turma.fk_dependencia_adm_id = dependencia_adm.pk_dependencia_adm_id')
        .where('turma.fk_tipo_turma_id <= 3')
        .where(`turma.ano_censo IN (${schoolClassYearQry})`)
        .group('turma.ano_censo')
        .group('dependencia_adm.nome')
        .group('estado.nome')
        .order('estado.nome')
        .order('dependencia_adm.nome');

    const enrollmentsPerSchoolLevelQry = squel.select()
        .field('estado.nome', 'name')
        .field('COALESCE(SUM(turma.num_matriculas), 0)', 'total')
        .field('turma.ano_censo', 'census_year')
        .field('etapa_ensino.desc_etapa', 'school_level')
        .from('turma')
        .from('etapa_ensino')
        .from('estado')
        .where('turma.fk_estado_id = estado.pk_estado_id')
        .where(`turma.fk_estado_id = ${stateId}`)
        .where('turma.fk_etapa_ensino_id = etapa_ensino.pk_etapa_ensino_id')
        .where('turma.fk_tipo_turma_id <= 3')
        .where(`turma.ano_censo IN (${schoolClassYearQry})`)
        .group('turma.ano_censo')
        .group('etapa_ensino.desc_etapa')
        .group('etapa_ensino.pk_etapa_ensino_id')
        .group('estado.nome')
        .order('estado.nome')
        .order('etapa_ensino.pk_etapa_ensino_id');

    const queryLabels = [ "school", "school_per_location", "enrollment", "enrollment_per_adm_dep",
        "enrollment_per_school_level" ];
    const querySet = [ totalSchoolsQry, schoolsPerLocationQry, enrollmentsQry,
        enrollmentsPerAdmDepQry, enrollmentsPerSchoolLevelQry];
    // wait until all queries finish or one of them fail
    Promise.all(dbExecAll(querySet)).then((queryResults) => {
        req.result = processResultSet(queryResults, queryLabels);
        for(let label in req.result) {
            for(let row of req.result[label]) {
                if (row.hasOwnProperty('location')) {
                    row.location = locationIdToStr(row.location);
                }
            }
        }
        next();
    }).catch((error) => {
        log.error(`[SQL query error] ${error}`);
        next(error);
    });
}, response('location'));

locationApp.get('/educational/city/:id', (req, res, next) => {
    const cityId = parseInt(req.params.id, 10);

    const censusYearQry = squel.select()
        .field('MAX(escola.ano_censo)', 'ano_censo')
        .from('escola')
        .toString();

    const totalSchoolsQry = squel.select()
        .field('municipio.nome', 'name')
        .field('0', 'location')
        .field('COUNT(DISTINCT(escola.cod_entidade))', 'total')
        .field('escola.ano_censo', 'census_year')
        .from('escola')
        .from('municipio')
        .where('escola.fk_municipio_id = municipio.pk_cod_ibge')
        .where(`escola.fk_municipio_id = ${cityId}`)
        .where(`escola.ano_censo IN (${censusYearQry})`)
        .where('escola.id_tipo_turma = 0')
        .group('municipio.nome')
        .group('escola.ano_censo')
        .order('municipio.nome');

    const schoolsPerLocationQry = squel.select()
        .field('municipio.nome', 'name')
        .field('escola.tipo_localizacao', 'location')
        .field('COUNT(DISTINCT(escola.cod_entidade))', 'total')
        .field('escola.ano_censo', 'census_year')
        .from('escola')
        .from('municipio')
        .where('escola.fk_municipio_id = municipio.pk_cod_ibge')
        .where(`escola.fk_municipio_id = ${cityId}`)
        .where(`escola.ano_censo IN (${censusYearQry})`)
        .where('escola.id_tipo_turma = 0')
        .group('municipio.nome')
        .group('escola.tipo_localizacao')
        .group('escola.ano_censo')
        .order('municipio.nome')
        .order('escola.tipo_localizacao');

    const schoolClassYearQry = squel.select()
        .field('MAX(turma.ano_censo)')
        .from('turma')
        .toString();

    const enrollmentsQry = squel.select()
        .field('municipio.nome', 'name')
        .field('COALESCE(SUM(turma.num_matriculas), 0)', 'total')
        .field('turma.ano_censo', 'census_year')
        .from('turma')
        .from('municipio')
        .where('turma.fk_municipio_id = municipio.pk_cod_ibge')
        .where(`turma.fk_municipio_id = ${cityId}`)
        .where(`turma.ano_censo IN (${schoolClassYearQry})`)
        .where('turma.fk_tipo_turma_id <= 3')
        .group('turma.ano_censo')
        .group('municipio.nome')
        .order('municipio.nome');

    const enrollmentsPerAdmDepQry = squel.select()
        .field('municipio.nome', 'name')
        .field('COALESCE(SUM(turma.num_matriculas), 0)', 'total')
        .field('turma.ano_censo', 'census_year')
        .field('dependencia_adm.nome', 'adm_dependency')
        .from('turma')
        .from('dependencia_adm')
        .from('municipio')
        .where('turma.fk_municipio_id = municipio.pk_cod_ibge')
        .where(`turma.fk_municipio_id = ${cityId}`)
        .where('turma.fk_dependencia_adm_id = dependencia_adm.pk_dependencia_adm_id')
        .where('turma.fk_tipo_turma_id <= 3')
        .where(`turma.ano_censo IN (${schoolClassYearQry})`)
        .group('turma.ano_censo')
        .group('dependencia_adm.nome')
        .group('municipio.nome')
        .order('municipio.nome')
        .order('dependencia_adm.nome');

    const enrollmentsPerSchoolLevelQry = squel.select()
        .field('municipio.nome', 'name')
        .field('COALESCE(SUM(turma.num_matriculas), 0)', 'total')
        .field('turma.ano_censo', 'census_year')
        .field('etapa_ensino.desc_etapa', 'school_level')
        .from('turma')
        .from('etapa_ensino')
        .from('municipio')
        .where('turma.fk_municipio_id = municipio.pk_cod_ibge')
        .where(`turma.fk_municipio_id = ${cityId}`)
        .where('turma.fk_etapa_ensino_id = etapa_ensino.pk_etapa_ensino_id')
        .where('turma.fk_tipo_turma_id <= 3')
        .where(`turma.ano_censo IN (${schoolClassYearQry})`)
        .group('turma.ano_censo')
        .group('etapa_ensino.desc_etapa')
        .group('etapa_ensino.pk_etapa_ensino_id')
        .group('municipio.nome')
        .order('municipio.nome')
        .order('etapa_ensino.pk_etapa_ensino_id');

    const queryLabels = [ "school", "school_per_location", "enrollment", "enrollment_per_adm_dep",
        "enrollment_per_school_level" ];
    const querySet = [ totalSchoolsQry, schoolsPerLocationQry, enrollmentsQry,
        enrollmentsPerAdmDepQry, enrollmentsPerSchoolLevelQry];
    // wait until all queries finish or one of them fail
    Promise.all(dbExecAll(querySet)).then((queryResults) => {
        req.result = processResultSet(queryResults, queryLabels);
        for(let label in req.result) {
            for(let row of req.result[label]) {
                if (row.hasOwnProperty('location')) {
                    row.location = locationIdToStr(row.location);
                }
            }
        }
        next();
    }).catch((error) => {
        log.error(`[SQL query error] ${error}`);
        next(error);
    });
}, response('location'));

locationApp.get('/educational/school_level', (req, res, next) => {
    const enrollmentsPerSchoolLevelYearQry = squel.select()
        .field('MAX(turma.ano_censo)', 'census_year')
        .from('turma');

    const enrollmentsPerSchoolLevelQry = squel.select()
        .field('COALESCE(SUM(turma.num_matriculas), 0)', 'total')
        .field('turma.ano_censo', 'census_year')
        .field('turma.serie_ano', 'school_year')
        .field('etapa_ensino.desc_etapa', 'school_level')
        .from('turma')
        .from('etapa_ensino')
        .where(`turma.ano_censo IN (${enrollmentsPerSchoolLevelYearQry.toString()})`)
        .where('turma.fk_etapa_ensino_id = etapa_ensino.pk_etapa_ensino_id')
        .where('turma.fk_tipo_turma_id <= 3')
        .group('etapa_ensino.desc_etapa')
        .group('etapa_ensino.pk_etapa_ensino_id')
        .group('turma.serie_ano')
        .group('turma.ano_censo')
        .order('etapa_ensino.pk_etapa_ensino_id')
        .order('turma.serie_ano')
        .order('turma.ano_censo');

    const queryLabels = [ 'enrollment_per_school_level', 'enrollment_census_year' ];
    const querySet = [ enrollmentsPerSchoolLevelQry, enrollmentsPerSchoolLevelYearQry ];
    // wait until all queries finish or one of them fail
    Promise.all(dbExecAll(querySet, enrollmentsPerSchoolLevelYearQry)).then((queryResults) => {
        const result = queryResults[0];
        const censusYear = queryResults[1][0]['census_year'];

        let school_levels = {};
        for(let i = 0; i < result.length; ++i) {
            const school_year  = schoolYearIdToStr(result[i].school_year);
            const school_level = result[i].school_level;
            const census_year = result[i].census_year;
            if (typeof school_levels[school_level] === 'undefined') {
                school_levels[school_level] = {};
            }
            school_levels[school_level][school_year] = parseInt(result[i].total, 10);
        }

        let response = [];
        for(let level in school_levels) {
            if (school_levels.hasOwnProperty(level)) {
                let sclevel = {};
                sclevel["degree"] = level;
                sclevel["census_year"] = parseInt(censusYear, 10);
                sclevel["table"] = [];
                for(let school_year in school_levels[level]) {
                    if (school_levels[level].hasOwnProperty(school_year)) {
                        let enrollment = { 'title' : school_year,
                                           'value' : school_levels[level][school_year] };
                        sclevel["table"].push(enrollment);
                    }
                }
                response.push(sclevel);
            }
        }
        req.result = response;
        next();
    }).catch((error) => {
        log.error(`[SQL query error] ${error}`);
        next(error);
    });
}, response('location'));

locationApp.get('/educational/school_level/region/:id', (req, res, next) => {
    const regionId = parseInt(req.params.id, 10);

    const enrollmentsPerSchoolLevelYearQry = squel.select()
        .field('MAX(turma.ano_censo)', 'census_year')
        .from('turma');

    const enrollmentsPerSchoolLevelQry = squel.select()
        .field('COALESCE(SUM(turma.num_matriculas), 0)', 'total')
        .field('turma.ano_censo', 'census_year')
        .field('turma.serie_ano', 'school_year')
        .field('etapa_ensino.desc_etapa', 'school_level')
        .from('turma')
        .from('etapa_ensino')
        .from('regiao')
        .where(`turma.fk_regiao_id = ${regionId}`)
        .where('turma.fk_regiao_id = regiao.pk_regiao_id')
        .where(`turma.ano_censo IN (${enrollmentsPerSchoolLevelYearQry.toString()})`)
        .where('turma.fk_etapa_ensino_id = etapa_ensino.pk_etapa_ensino_id')
        .where('turma.fk_tipo_turma_id <= 3')
        .group('regiao.nome')
        .group('etapa_ensino.desc_etapa')
        .group('etapa_ensino.pk_etapa_ensino_id')
        .group('turma.serie_ano')
        .group('turma.ano_censo')
        .order('regiao.nome')
        .order('etapa_ensino.pk_etapa_ensino_id')
        .order('turma.serie_ano')
        .order('turma.ano_censo');

    const queryLabels = [ 'enrollment_per_school_level', 'enrollment_census_year' ];
    const querySet = [ enrollmentsPerSchoolLevelQry, enrollmentsPerSchoolLevelYearQry ];
    // wait until all queries finish or one of them fail
    Promise.all(dbExecAll(querySet, enrollmentsPerSchoolLevelYearQry)).then((queryResults) => {
        const result = queryResults[0];
        const censusYear = queryResults[1][0]['census_year'];

        let school_levels = {};
        for(let i = 0; i < result.length; ++i) {
            const school_year  = schoolYearIdToStr(result[i].school_year);
            const school_level = result[i].school_level;
            const census_year = result[i].census_year;
            if (typeof school_levels[school_level] === 'undefined') {
                school_levels[school_level] = {};
            }
            school_levels[school_level][school_year] = parseInt(result[i].total, 10);
        }

        let response = [];
        for(let level in school_levels) {
            if (school_levels.hasOwnProperty(level)) {
                let sclevel = {};
                sclevel["degree"] = level;
                sclevel["census_year"] = parseInt(censusYear, 10);
                sclevel["table"] = [];
                for(let school_year in school_levels[level]) {
                    if (school_levels[level].hasOwnProperty(school_year)) {
                        let enrollment = { 'title' : school_year,
                                           'value' : school_levels[level][school_year] };
                        sclevel["table"].push(enrollment);
                    }
                }
                response.push(sclevel);
            }
        }
        req.result = response;
        next();
    }).catch((error) => {
        log.error(`[SQL query error] ${error}`);
        next(error);
    });
}, response('location'));

locationApp.get('/educational/school_level/state/:id', (req, res, next) => {
    const stateId = parseInt(req.params.id, 10);

    const enrollmentsPerSchoolLevelYearQry = squel.select()
        .field('MAX(turma.ano_censo)', 'census_year')
        .from('turma');

    const enrollmentsPerSchoolLevelQry = squel.select()
        .field('COALESCE(SUM(turma.num_matriculas), 0)', 'total')
        .field('turma.ano_censo', 'census_year')
        .field('turma.serie_ano', 'school_year')
        .field('etapa_ensino.desc_etapa', 'school_level')
        .from('turma')
        .from('etapa_ensino')
        .from('estado')
        .where(`turma.fk_estado_id = ${stateId}`)
        .where('turma.fk_estado_id = estado.pk_estado_id')
        .where(`turma.ano_censo IN (${enrollmentsPerSchoolLevelYearQry.toString()})`)
        .where('turma.fk_etapa_ensino_id = etapa_ensino.pk_etapa_ensino_id')
        .where('turma.fk_tipo_turma_id <= 3')
        .group('estado.nome')
        .group('etapa_ensino.desc_etapa')
        .group('etapa_ensino.pk_etapa_ensino_id')
        .group('turma.serie_ano')
        .group('turma.ano_censo')
        .order('estado.nome')
        .order('etapa_ensino.pk_etapa_ensino_id')
        .order('turma.serie_ano')
        .order('turma.ano_censo');

    const queryLabels = [ 'enrollment_per_school_level', 'enrollment_census_year' ];
    const querySet = [ enrollmentsPerSchoolLevelQry, enrollmentsPerSchoolLevelYearQry ];
    // wait until all queries finish or one of them fail
    Promise.all(dbExecAll(querySet, enrollmentsPerSchoolLevelYearQry)).then((queryResults) => {
        const result = queryResults[0];
        const censusYear = queryResults[1][0]['census_year'];

        let school_levels = {};
        for(let i = 0; i < result.length; ++i) {
            const school_year  = schoolYearIdToStr(result[i].school_year);
            const school_level = result[i].school_level;
            const census_year = result[i].census_year;
            if (typeof school_levels[school_level] === 'undefined') {
                school_levels[school_level] = {};
            }
            school_levels[school_level][school_year] = parseInt(result[i].total, 10);
        }

        let response = [];
        for(let level in school_levels) {
            if (school_levels.hasOwnProperty(level)) {
                let sclevel = {};
                sclevel["degree"] = level;
                sclevel["census_year"] = parseInt(censusYear, 10);
                sclevel["table"] = [];
                for(let school_year in school_levels[level]) {
                    if (school_levels[level].hasOwnProperty(school_year)) {
                        let enrollment = { 'title' : school_year,
                                           'value' : school_levels[level][school_year] };
                        sclevel["table"].push(enrollment);
                    }
                }
                response.push(sclevel);
            }
        }
        req.result = response;
        next();
    }).catch((error) => {
        log.error(`[SQL query error] ${error}`);
        next(error);
    });
}, response('location'));

locationApp.get('/educational/school_level/city/:id', (req, res, next) => {
    const cityId = parseInt(req.params.id, 10);

    const enrollmentsPerSchoolLevelYearQry = squel.select()
        .field('MAX(turma.ano_censo)', 'census_year')
        .from('turma');

    const enrollmentsPerSchoolLevelQry = squel.select()
        .field('COALESCE(SUM(turma.num_matriculas), 0)', 'total')
        .field('turma.ano_censo', 'census_year')
        .field('turma.serie_ano', 'school_year')
        .field('etapa_ensino.desc_etapa', 'school_level')
        .from('turma')
        .from('etapa_ensino')
        .from('municipio')
        .where(`turma.fk_municipio_id = ${cityId}`)
        .where('turma.fk_municipio_id = municipio.pk_cod_ibge')
        .where(`turma.ano_censo IN (${enrollmentsPerSchoolLevelYearQry.toString()})`)
        .where('turma.fk_etapa_ensino_id = etapa_ensino.pk_etapa_ensino_id')
        .where('turma.fk_tipo_turma_id <= 3')
        .group('municipio.nome')
        .group('etapa_ensino.desc_etapa')
        .group('etapa_ensino.pk_etapa_ensino_id')
        .group('turma.serie_ano')
        .group('turma.ano_censo')
        .order('municipio.nome')
        .order('etapa_ensino.pk_etapa_ensino_id')
        .order('turma.serie_ano')
        .order('turma.ano_censo');

    const queryLabels = [ 'enrollment_per_school_level', 'enrollment_census_year' ];
    const querySet = [ enrollmentsPerSchoolLevelQry, enrollmentsPerSchoolLevelYearQry ];
    // wait until all queries finish or one of them fail
    Promise.all(dbExecAll(querySet, enrollmentsPerSchoolLevelYearQry)).then((queryResults) => {
        const result = queryResults[0];
        const censusYear = queryResults[1][0]['census_year'];

        let school_levels = {};
        for(let i = 0; i < result.length; ++i) {
            const school_year  = schoolYearIdToStr(result[i].school_year);
            const school_level = result[i].school_level;
            const census_year = result[i].census_year;
            if (typeof school_levels[school_level] === 'undefined') {
                school_levels[school_level] = {};
            }
            school_levels[school_level][school_year] = parseInt(result[i].total, 10);
        }

        let response = [];
        for(let level in school_levels) {
            if (school_levels.hasOwnProperty(level)) {
                let sclevel = {};
                sclevel["degree"] = level;
                sclevel["census_year"] = parseInt(censusYear, 10);
                sclevel["table"] = [];
                for(let school_year in school_levels[level]) {
                    if (school_levels[level].hasOwnProperty(school_year)) {
                        let enrollment = { 'title' : school_year,
                                           'value' : school_levels[level][school_year] };
                        sclevel["table"].push(enrollment);
                    }
                }
                response.push(sclevel);
            }
        }
        req.result = response;
        next();
    }).catch((error) => {
        log.error(`[SQL query error] ${error}`);
        next(error);
    });
}, response('location'));

module.exports = locationApp;
